package net.sf.r4h;

import static net.sf.r4h.Restrictions.equal;

import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.Serializable;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.r4h.Restrictions.CaseAwareRestriction;
import net.sf.r4h.Restrictions.EmptinessAwareRestriction;
import net.sf.r4h.Restrictions.ExampleRestriction;

import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.PropertyAccessException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.criterion.Conjunction;
import org.hibernate.criterion.CriteriaQuery;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Example;
import org.hibernate.criterion.LikeExpression;
import org.hibernate.criterion.Projections;
import org.hibernate.criterion.SimpleExpression;
import org.hibernate.engine.spi.SessionFactoryImplementor;
import org.hibernate.internal.SessionFactoryImpl;
import org.hibernate.persister.entity.EntityPersister;
import org.hibernate.sql.JoinType;
import org.hibernate.tuple.StandardProperty;
import org.hibernate.type.CollectionType;
import org.hibernate.type.ManyToOneType;
import org.hibernate.type.Type;

/**
 * /**
 * 
 * The goal of this class is to provide API that allows to write common CRUD operations in ONE line of code.
 * 
 * In addition to it's methods it provides paging, improved ordering (see {@link Order}) and restriction (see {@link Restrictions})
 * 
 * In methods with paging pages numbers start from 1. By default Restriction is {@link Restrictions#EQUAL}
 * 
 * @author vitaliy.se@gmail.com
 * @since 1.0.0
 * @version 4.0.0
 * 
 *
 * @param <E> a type of entity 
 * @param <I> a type of identity key 
 */
public class HibernateRepository<E, I extends Serializable> {

	private static final Partition ALL_IN_ONE_PARTITION = new Partition(0,Integer.MAX_VALUE);
	private final SessionFactory sessionFactory;
	private final Class<E> entityClass;

	/**
	 * 
	 * @param sessionFactory - a Hibernate's SessionFactory which implements org.hibernate.engine.spi.SessionFactoryImplementor
	 * @param entityClass - an entity class for repository
	 */
	public HibernateRepository(SessionFactory sessionFactory, Class<E> entityClass) {
		if (sessionFactory!=null && !(sessionFactory instanceof SessionFactoryImpl)) {
			throw new IllegalArgumentException("SessionFactory is not instance of " + SessionFactoryImpl.class.getCanonicalName());
		}
		this.sessionFactory = sessionFactory;
		this.entityClass = entityClass;
	}

	public List<E> list(ExampleRestriction restriction, E exampleObject, Page page, Order... orderStrategies) {
		return list("", restriction, exampleObject, pageToPartition(page), orderStrategies);
	}
	
	public List<E> list(ExampleRestriction restriction, E exampleObject, Partition partition, Order... orderStrategies) {
		return list("", restriction, exampleObject, partition, orderStrategies);
	}
	

	public List<E> list(E exampleObject, Page page, Order... orderStrategies) {
		return list(Restrictions.example(), exampleObject, pageToPartition(page), orderStrategies);
	}
	
	public List<E> list(E exampleObject, Partition partition, Order... orderStrategies) {
		return list(Restrictions.example(), exampleObject, partition, orderStrategies);
	}

	@SuppressWarnings("unchecked")
	public List<E> list(ExampleRestriction restriction, Object exampleObject, Order... orderStrategies) {
		return list(restriction, (E) exampleObject, ALL_IN_ONE_PARTITION, orderStrategies);
	}

	public List<E> list(E exampleObject, Order... orderStrategies) {
		return list(Restrictions.example(), exampleObject, orderStrategies);
	}

	public Long count(ExampleRestriction restriction, E exampleObject) {
		return count("", restriction, exampleObject);
	}

	public Long count(E exampleObject) {
		return count(Restrictions.example(), exampleObject);
	}

	public List<E> list(String[] properties, Restrictions.Restriction[] restrictions, Object[] values, Page page, Order... orderStrategies) {
		return execute(new CriteriaHibernateCallback<E>(entityClass, properties, restrictions, values, pageToPartition(page), getSessionFactoryImplementor()).addOrder(orderStrategies));
	}
	
	public List<E> list(String[] properties, Restrictions.Restriction[] restrictions, Object[] values, Partition partition, Order... orderStrategies) {
		return execute(new CriteriaHibernateCallback<E>(entityClass, properties, restrictions, values, partition, getSessionFactoryImplementor()).addOrder(orderStrategies));
	}

	public List<E> list(String[] properties, Restrictions.Restriction[] restrictions, Object[] values, Order... orderStrategies) {
		return list(properties, restrictions, values, ALL_IN_ONE_PARTITION, orderStrategies);
	}

	public Long count(String[] properties, Restrictions.Restriction[] restrictions, Object[] values) {
		return execute(new CriteriaHibernateCallback<E>(entityClass, properties, restrictions, values, getSessionFactoryImplementor()).setReturnCount());
	}

	public List<E> list(String property, Restrictions.Restriction restriction, Object value, Page page, Order... orderStrategies) {
		return list(new String[] { property }, new Restrictions.Restriction[] { restriction }, new Object[] { value }, pageToPartition(page), orderStrategies);
	}
	
	public List<E> list(String property, Restrictions.Restriction restriction, Object value, Partition partition, Order... orderStrategies) {
		return list(new String[] { property }, new Restrictions.Restriction[] { restriction }, new Object[] { value }, partition, orderStrategies);
	}

	public List<E> list(String property, Restrictions.Restriction restriction, Object value, Order... orderStrategies) {
		return list(new String[] { property }, new Restrictions.Restriction[] { restriction }, new Object[] { value }, orderStrategies);
	}

	public Long count(String property, Restrictions.Restriction restriction, Object value) {
		return count(new String[] { property }, new Restrictions.Restriction[] { restriction }, new Object[] { value });
	}

	public List<E> list(String property, Object value, Page page, Order... orderStrategies) {
		return list(property, equal(), value, pageToPartition(page), orderStrategies);
	}
	
	public List<E> list(String property, Object value, Partition partition, Order... orderStrategies) {
		return list(property, equal(), value, partition, orderStrategies);
	}


	public List<E> list(String property, Object value, Order... orderStrategies) {
		return list(property, value, ALL_IN_ONE_PARTITION, orderStrategies);
	}

	public Long count(String property, Object value) {
		return count(property, equal(), value);
	}

	public List<E> list(String[] properties, Object[] values, Page page, Order... orderStrategies) {
		return list(properties, HibernateRepository.createArrayOfEqualRestrictions(properties.length), values, pageToPartition(page), orderStrategies);
	}
	
	public List<E> list(String[] properties, Object[] values, Partition partition, Order... orderStrategies) {
		return list(properties, HibernateRepository.createArrayOfEqualRestrictions(properties.length), values, partition, orderStrategies);
	}

	public List<E> list(String[] properties, Object[] values, Order... orderStrategies) {
		return list(properties, HibernateRepository.createArrayOfEqualRestrictions(properties.length), values, orderStrategies);
	}

	public Long count(String[] properties, Object[] values) {
		return count(properties, HibernateRepository.createArrayOfEqualRestrictions(properties.length), values);
	}

	public List<E> list(Page page, Order... orderStrategies) {
		return list( HibernateRepositoryUtils.properties(), HibernateRepositoryUtils.values(), pageToPartition(page), orderStrategies);
	}
	
	public List<E> list(Partition partition, Order... orderStrategies) {
		return list( HibernateRepositoryUtils.properties(), HibernateRepositoryUtils.values(), partition, orderStrategies);
	}
	

	public List<E> list(Order... orderStrategies) {
		return list( ALL_IN_ONE_PARTITION, orderStrategies);
	}

	/**
	 * @return count of all entities of the entityType specified in constructor 
	 */
	public Long count() {
		return execute(new CriteriaHibernateCallback<E>(entityClass).setReturnCount());
	}
	

	@SuppressWarnings("unchecked")
	public E get(I id) {
		return (E) sessionFactory.getCurrentSession().get(entityClass, id);
	}

	public boolean contains(E entity) {
		return sessionFactory.getCurrentSession().contains(entity);
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> listByHql(String hql, String[] namedParams, Object[] values) {
		Query createQuery = sessionFactory.getCurrentSession().createQuery(hql);
		for (int i = 0; i < values.length; i++) {
			createQuery.setParameter(namedParams[i], values[i]);
		}
		return createQuery.list();
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> listByHql(String hql) {
		return sessionFactory.getCurrentSession().createQuery(hql).list();
	}

	@SuppressWarnings("unchecked")
	public <T> List<T> listByHql(String hql, Object... values) {
		Query createQuery = sessionFactory.getCurrentSession().createQuery(hql);
		for (int i = 0; i < values.length; i++) {
			createQuery.setParameter(i, values[i]);
		}
		return createQuery.list();
	}

	@SuppressWarnings("unchecked")
	public E unique(String[] properties, Restrictions.Restriction[] restrictions, Object[] values) {
		// this cast fixes the compiler error: Restriction parameters of <T>T cannot be determined; no unique maximal instance exists for Restriction variable T with upper bounds T,java.lang.Object
		return (E) execute(new CriteriaHibernateCallback<E>(entityClass, properties, restrictions, values, getSessionFactoryImplementor()).setUnique());
	}

	public E unique(String[] properties, Object[] values) {
		return unique(properties, HibernateRepository.createArrayOfEqualRestrictions(properties.length), values);
	}

	public E unique(String propertyName, Restrictions.Restriction restrictions, Object propertyValue) {
		return unique(new String[] { propertyName }, HibernateRepositoryUtils.restrictions(restrictions), new Object[] { propertyValue });
	}

	public E unique(String propertyName, Object propertyValue) {
		return unique(propertyName, equal(), propertyValue);
	}


	public Session getCurrentSession() {
		return sessionFactory.getCurrentSession();
	}

	@SuppressWarnings("unchecked")
	public E load(I id) {
		return (E) getCurrentSession().load(entityClass, id);
	}

	@SuppressWarnings("unchecked")
	public List<E> loadAll() {
		return getCurrentSession().createCriteria(entityClass).list();
	}

	public E saveOrUpdate(E entity) {
		getCurrentSession().saveOrUpdate(entity);
		return entity;
	}

	/**
	 * Saves or updates entity and flushes session, this allows to use saved entity id
	 * 
	 * @param entity
	 * @return saves entity
	 */
	public E saveOrUpdateAndFlush(E entity) {
		entity = saveOrUpdate(entity);
		getCurrentSession().flush();
		return entity;
	}

	public E save(E entity) {
		getCurrentSession().save(entity);
		return entity;
	}

	/**
	 * Saves entity and flushes session, this allows to use saved entity id
	 * 
	 * @param entity
	 * @return saves entity
	 */
	public E saveAndFlush(E entity) {
		entity = save(entity);
		getCurrentSession().flush();
		return entity;
	}

	public E merge(E entity) {
		getCurrentSession().merge(entity);
		return entity;
	}

	public E delete(E entity) {
		getCurrentSession().delete(entity);
		return entity;
	}

	public E delete(Serializable id) {
		@SuppressWarnings("unchecked")
		E entity = (E) getCurrentSession().load(entityClass, id);
		getCurrentSession().delete(entity);
		return entity;
	}

	/**
	 * Deletes all entities of given entityName
	 * 
	 * @param entityName
	 * @return number of deleted entities
	 */
	public int deleteAll(String entityName) {
		return getCurrentSession().createQuery("delete from " + entityName).executeUpdate();
	}

	/**
	 * Deletes all entities of given entityClass
	 * 
	 * @return number of deleted entities
	 */
	public int deleteAll() {
		return deleteAll(entityClass.getName());
	}
	/**
	 * First fetches all entities and than deletes it one by one.
	 * Is needed to correctly delete graphs of objects since it deletes references and unlike {@link HibernateRepository#deleteAll()} avoids constraint violation 
	 * WARNING: This method is useful in tests, but seldom needed on production since it is very inefficient. 
	 * 
	 * @return
	 */
	public List<E> deleteAllPreFetched() {
		List<E> entities = list();
		for (E e : entities) {
			delete(e);
		}
		return entities;
	}

	@SuppressWarnings("unchecked")
	protected <T> T execute(CriteriaHibernateCallback<?> criteriaHibernateCallback) {
		return (T) criteriaHibernateCallback.doInHibernate(getCurrentSession());
	}
	
	private SessionFactoryImpl getSessionFactoryImplementor() {
		return (SessionFactoryImpl) sessionFactory;
	}

	private static Restrictions.Restriction[] createArrayOfEqualRestrictions(int size) {
		Restrictions.Restriction[] restrictions = new Restrictions.Restriction[size];
		Arrays.fill(restrictions, equal());
		return restrictions;
	}

	/**
	 * HibernateCallback using Hibernate criteria API for setting restriction and ordering.
	 * 
	 * Used in {@link HibernateRepository} and by brave souls who is not afraid to dig into Hibernate's source code.
	 * 
	 * @param <T> entity for query
	 * 
	 * @author vitaliy.se@gmail.com
	 * @since 1.0.0
	 */
	private static class CriteriaHibernateCallback<T> {

		private String[] properties;
		private Object[] values;
		private Restrictions.Restriction[] restrictions;
		private final Partition partition;
		private SessionFactoryImpl implementor;
		private List<Order> orders = new ArrayList<Order>();
		private Class<?> entityClass;
		private boolean isUnique;
		private boolean returnCount = false;
		private boolean returnNothing;

		/**
		 * 
		 * @param entityClass the type to return
		 * @param properties properties for restrictions to be applied
		 * @param restrictions
		 * @param values
		 * @param partition
		 * @param pageSize
		 * @param implementor
		 */
		public CriteriaHibernateCallback(Class<?> entityClass, String properties[], Restrictions.Restriction[] restrictions, Object values[], Partition partition, SessionFactoryImpl implementor) {
			this.entityClass = entityClass;
			this.partition = partition;
			this.implementor = implementor;
			this.properties = properties;
			this.values = values;
			this.restrictions = restrictions;
		}

		public CriteriaHibernateCallback(Class<?> entityClass, String properties[], Restrictions.Restriction[] restrictions, Object values[], SessionFactoryImpl implementor) {
			this(entityClass, properties, restrictions, values, ALL_IN_ONE_PARTITION, implementor);
		}

		public CriteriaHibernateCallback(Class<?> entityClass) {
			this(entityClass, new String[] {}, new Restrictions.Restriction[] {}, new Object[] {}, ALL_IN_ONE_PARTITION, null);
		}

		public boolean isUnique() {
			return isUnique;
		}

		/**
		 * @return true if it is a count query
		 */
		public boolean isReturnCount() {
			return returnCount;
		}

		public CriteriaHibernateCallback<T> setReturnCount() {
			this.returnCount = true;
			return this;
		}

		public CriteriaHibernateCallback<T> setUnique() {
			this.isUnique = true;
			return this;
		}

		/**
		 * 
		 * @param orders for query
		 * @return {@link CriteriaHibernateCallback}
		 */
		public CriteriaHibernateCallback<T> addOrder(Order... orders) {
			this.orders.addAll(Arrays.asList(orders));
			return this;
		}

		public final T doInHibernate(Session session) {
			session.flush();// FIXME fix for http://opensource.atlassian.com/projects/hibernate/browse/HHH-3813
			Criteria criteria = session.createCriteria(entityClass);
			Map<String, Criteria> assosiationPathCriteriaMap = new HashMap<String, Criteria>();
			applyRestriction(criteria, assosiationPathCriteriaMap);
			if (!returnNothing) {
				if (isReturnCount()) {
					criteria.setProjection(Projections.rowCount());
				} else {
					for (Order order : orders) {
						// TODO find out why it was commented
						// creationAssociation(criteria, assosiationPathCriteriaMap, order.getSortedFieldPath()).addOrder(createHibernateOrder(order, getLastPropertyName(order.getSortedFieldPath())));
						creationAssociation(criteria, assosiationPathCriteriaMap, order.getSortedFieldPath()).addOrder(new NullsAwareHibernateOrder(order));
					}
				}
			}
			return extractResults(criteria);
		}

		@SuppressWarnings("unchecked")
		private T extractResults(Criteria criteria) {
			if (isReturnCount() ) {
				return (T) criteria.uniqueResult();
			}else if  (isUnique()) {
				return returnNothing ? null : (T) criteria.uniqueResult();
			}
			return returnNothing ? (T) Collections.EMPTY_LIST : (T) criteria.list();
		}

		public void applyRestriction(Criteria criteria, Map<String, Criteria> assosiationPathCriteriaMap) {
			if (partition.getMax()== 0) {
				returnNothing = true;// there is no need to proceed;
				return;// TODO Consider throwing IllegalArugmentException in case the pageSise is Negative
			}
			criteria.setFirstResult(partition.getFirst()).setMaxResults(partition.getMax());

			for (int i = 0; i < properties.length; i++) {
				Object value = i < values.length ? values[i] : null; // allows not to specify values in the end if they are null
				boolean isEmptyCollectionOrArray = ((value instanceof Collection<?>) && ((Collection<?>) value).isEmpty()) || ((value instanceof Object[]) && ((Object[]) value).length == 0);
				if (restrictions[i] instanceof Restrictions.ExampleRestriction) {
					criteria.add(createExample(value, (ExampleRestriction) restrictions[i]));
					if (properties[i].length() == 0) {
						addAssosiationFilteringByExample(value, criteria, assosiationPathCriteriaMap, "", new HashMap<Class<?>, Set<String>>(), (ExampleRestriction) restrictions[i]);
					} else {
						// TODO test this approach
						Criteria subCriteria = creationAssociation(criteria, assosiationPathCriteriaMap, properties[i]);
						addAssosiationFilteringByExample(value, subCriteria, assosiationPathCriteriaMap, properties[i], new HashMap<Class<?>, Set<String>>(), (ExampleRestriction) restrictions[i]);
					}
				} else if (value != null && !(restrictions[i] instanceof EmptinessAwareRestriction && ((EmptinessAwareRestriction) restrictions[i]).isDisabledIfValueIsEmpty()) && isEmptyCollectionOrArray) {
					returnNothing = true;// IN Statment with empty list, there is no need to proceed cause we do not have OR conditions
					criteria.add(org.hibernate.criterion.Restrictions.sqlRestriction("1=0"));//needed for count to return 0
					return;
				} else if (!((restrictions[i].isDisabledIfValueIsNull() && value == null) || (restrictions[i] instanceof EmptinessAwareRestriction) && ((EmptinessAwareRestriction) restrictions[i]).isDisabledIfValueIsEmpty() && isEmptyCollectionOrArray)) {
					Criteria subCriteria = creationAssociation(criteria, assosiationPathCriteriaMap, properties[i]);
					Criterion criterion = createCriterion(getLastPropertyName(properties[i]), restrictions[i], value);
					if (restrictions[i] instanceof CaseAwareRestriction && !((CaseAwareRestriction) restrictions[i]).isCaseSensitive()) {
						((SimpleExpression) criterion).ignoreCase();
					}
					subCriteria.add(criterion);
				} else { // i.e. this is either isDisabledIfEmpty or isDisabledIfValueIsNull
					// do nothing
				}
			}
		}

		private Criterion createCriterion(String property, Restrictions.Restriction restriction, Object value) {
			switch (restriction.getType()) {
			case LESS:
				return org.hibernate.criterion.Restrictions.lt(property, value);
			case LESS_OR_EQUAL:
				return org.hibernate.criterion.Restrictions.le(property, value);
			case EQUAL:
				return org.hibernate.criterion.Restrictions.eq(property, value);
			case NOT_EQUAL:
				return org.hibernate.criterion.Restrictions.ne(property, value);
			case GREATER_OR_EQUAL:
				return org.hibernate.criterion.Restrictions.ge(property, value);
			case GREATER:
				return org.hibernate.criterion.Restrictions.gt(property, value);
			case LIKE:
				return org.hibernate.criterion.Restrictions.like(property, value);
			case ILIKE:
				//TODO revert once this bug is closed https://hibernate.onjira.com/browse/HHH-6915
				return new ILikeExpression(property, value.toString());
//				 return org.hibernate.criterion.Restrictions.ilike(property, value);
			case IN:
				if (value instanceof Collection<?>) {
					return org.hibernate.criterion.Restrictions.in(property, (Collection<?>) value);
				} else {
					return org.hibernate.criterion.Restrictions.in(property, (Object[]) value);
				}
			case NOT_IN:
				return org.hibernate.criterion.Restrictions.not(createCriterion(property,Restrictions.in(),value));
			case IS_EMPTY:
				return org.hibernate.criterion.Restrictions.isEmpty(property);
			case NOT_EMPTY:
				return org.hibernate.criterion.Restrictions.isNotEmpty(property);
			case IS_NULL:
				return org.hibernate.criterion.Restrictions.isNull(property);
			case NOT_NULL:
				return org.hibernate.criterion.Restrictions.isNotNull(property);
			case SIZE_LESS:
				return org.hibernate.criterion.Restrictions.sizeLt(property, (Integer) value);
			case SIZE_LESS_OR_EQUAL:
				return org.hibernate.criterion.Restrictions.sizeLe(property, (Integer) value);
			case SIZE_EQUAL:
				return org.hibernate.criterion.Restrictions.sizeEq(property, (Integer) value);
			case SIZE_GREATER_OR_EQUAL:
				return org.hibernate.criterion.Restrictions.sizeGe(property, (Integer) value);
			case SIZE_GREATER:
				return org.hibernate.criterion.Restrictions.sizeGt(property, (Integer) value);
			case PROPERTY_LESS:
				return org.hibernate.criterion.Restrictions.ltProperty(property, (String) value);
			case PROPERTY_LESS_OR_EQUAL:
				return org.hibernate.criterion.Restrictions.leProperty(property, (String) value);
			case PROPERTY_EQUAL:
				return org.hibernate.criterion.Restrictions.eqProperty(property, (String) value);
			case PROPERTY_GREATER_OR_EQUAL:
				return org.hibernate.criterion.Restrictions.geProperty(property, (String) value);
			case PROPERTY_GREATER:
				return org.hibernate.criterion.Restrictions.gtProperty(property, (String) value);
			default:
				throw new IllegalArgumentException("Unknown restriction=" + restriction + " was passed to method");
			}
		}

		private void addAssosiationFilteringByExample(Object example, Criteria criteria, Map<String, Criteria> assosiationPathCriteriaMap, String previousAssosiationPath, Map<Class<?>, Set<String>> entityWalkedProperties, ExampleRestriction exampleRestriction) {
			Set<String> visitedProperties = entityWalkedProperties.get(example.getClass());
			if (visitedProperties == null) {
				visitedProperties = new HashSet<String>();
				entityWalkedProperties.put(example.getClass(), visitedProperties);
			}
			for (String entityName : implementor.getImplementors(example.getClass().getName())) {
				EntityPersister persister = implementor.getEntityPersister(entityName);
				for (StandardProperty property : persister.getEntityMetamodel().getProperties()) {
					if (visitedProperties.contains(property.getName())) {
						continue;
					}
					if (property.getType() instanceof CollectionType || property.getType() instanceof ManyToOneType) {
						visitedProperties.add(property.getName());
					}
					Object propertyValue = null;
					try {
						//we have to get rid of apache beanuits because it depends on common-logging
//						propertyValue = PropertyUtils.getProperty(example, property.getName());
						for (PropertyDescriptor propertyDescriptor: Introspector.getBeanInfo(example.getClass()).getPropertyDescriptors()){
							if (propertyDescriptor.getName().equals(property.getName())){
								propertyValue = propertyDescriptor.getReadMethod().invoke(example);
							}
						}
						
					} catch (Exception e) {
						// It is safe to catch Exception here because example and property are not null
						// This situation is unlikely to happen, because Hibernate disallows usage of such objects 
						String errorMessage = e.getMessage() + " probably entity " + entityName + " ManyToOne.class annotates a property that has not coresponding public get method";
						throw new PropertyAccessException(e, errorMessage, false, example.getClass(), property.getName());
					}
					if (propertyValue != null) {
						String newAssosiationPath = previousAssosiationPath + property.getName();
						Conjunction conjunction = org.hibernate.criterion.Restrictions.conjunction();
						if (property.getType() instanceof CollectionType) {
							@SuppressWarnings("unchecked")
							Collection<Object> collection = (Collection<Object>) propertyValue;
							if (!collection.isEmpty()) {
								Criteria subCriteria = criteria.createCriteria(property.getName(), property.getName(), JoinType.LEFT_OUTER_JOIN).add(conjunction);
								assosiationPathCriteriaMap.put(newAssosiationPath, subCriteria);
								addAssosiationFilteringByExample(collection.iterator().next(), subCriteria, assosiationPathCriteriaMap, newAssosiationPath + ".", entityWalkedProperties, exampleRestriction);
								for (Object exampleObject : collection) {
									conjunction.add(createExample(exampleObject, exampleRestriction));
								}
							}
						} else if (property.getType() instanceof ManyToOneType) {
							Criteria subCriteria = criteria.createCriteria(property.getName(), property.getName(), JoinType.LEFT_OUTER_JOIN).add(createExample(propertyValue, exampleRestriction));
							assosiationPathCriteriaMap.put(newAssosiationPath, subCriteria);
							addAssosiationFilteringByExample(propertyValue, subCriteria, assosiationPathCriteriaMap, newAssosiationPath + ".", entityWalkedProperties, exampleRestriction);

						}
					}

				}
			}
		}

		private static Example createExample(Object object, ExampleRestriction exampleRestriction) {
			Example example = Example.create(object);
			if (exampleRestriction.areZeroesExcluded()) {
				example.excludeZeroes();
			}
			if (exampleRestriction.isCaseIgnored()) {
				example.ignoreCase();
			}
			if (exampleRestriction.isLikeEnabled()) {
				example.enableLike();
			}
			if (exampleRestriction.isExcludeNone()) {
				example.excludeNone();
			}
			return example;
		}

		private static Criteria creationAssosiation(Criteria criteria, Map<String, Criteria> assosiationPathCriteriaMap, String fullAssosiationPath, String previosAssosiationPath) {
			Criteria subCriteria;
			int dotIndex = fullAssosiationPath.indexOf('.');
			if (dotIndex == -1) {
				subCriteria = criteria;
			} else {
				String propertyName = fullAssosiationPath.substring(0, dotIndex);
				String newAssosiationPath = previosAssosiationPath + propertyName;
				subCriteria = assosiationPathCriteriaMap.get(newAssosiationPath);
				if (subCriteria == null) {
					subCriteria = criteria.createCriteria(propertyName, propertyName, JoinType.LEFT_OUTER_JOIN);
					assosiationPathCriteriaMap.put(newAssosiationPath, subCriteria);
				}
				subCriteria = creationAssosiation(subCriteria, assosiationPathCriteriaMap, fullAssosiationPath.substring(dotIndex + 1), newAssosiationPath + ".");
			}
			return subCriteria;
		}

		/**
		 * 
		 * @param criteria
		 * @param associationPathCriteriaMap vman with key as association and criteria as value
		 * @param fullAssociationPath
		 * @return criteria with full association path for given fullAssociationPath
		 */
		private static Criteria creationAssociation(Criteria criteria, Map<String, Criteria> associationPathCriteriaMap, String fullAssociationPath) {
			return creationAssosiation(criteria, associationPathCriteriaMap, fullAssociationPath, "");
		}

		private static String getLastPropertyName(String fullAssosiationPath) {
			int dotIndex = fullAssosiationPath.lastIndexOf('.');
			if (dotIndex > -1) {
				return fullAssosiationPath.substring(dotIndex + 1);
			}
			return fullAssosiationPath;
		}

		private static class NullsAwareHibernateOrder extends org.hibernate.criterion.Order {
			private static final long serialVersionUID = 3353395213896284011L;
			private Order order;
			private String propertyName;

			private NullsAwareHibernateOrder(Order order) {
				// here we are passing useless parameters to super class
				super(order.getSortedFieldPath(), order.getSortingDirection() == Order.SortingDirection.ASCENDING ? true : false);
				this.order = order;
				propertyName = getLastPropertyName(order.getSortedFieldPath());
			}

			@Override
			public String toSqlString(Criteria criteria, CriteriaQuery criteriaQuery) throws HibernateException {
				// TODO consider if we should use here sortedFieldPath instead of propertyName
				String[] columns = criteriaQuery.getColumnsUsingProjection(criteria, propertyName);
				Type type = criteriaQuery.getTypeUsingProjection(criteria, propertyName);
				StringBuilder sql = new StringBuilder();
				for (int i = 0; i < columns.length; i++) {
					SessionFactoryImplementor factory = criteriaQuery.getFactory();
					boolean lower = order.getGetCaseSensitivity() == Order.CaseSensitivity.INSENSITIVE && type.sqlTypes(factory)[i] == Types.VARCHAR;
					if (lower) {
						sql.append(factory.getDialect().getLowercaseFunction()).append('(');
					}
					sql.append(columns[i]);
					if (lower) {
						sql.append(')');
					}
					sql.append(order.getSortingDirection() == Order.SortingDirection.ASCENDING ? " asc" : " desc");
					if (order.getNullsPosition() != Order.NullsPosition.NOT_SPECIFIED) {
						sql.append(order.getNullsPosition() == Order.NullsPosition.FIRST ? " nulls first" : " nulls last");
					}
					if (i < columns.length - 1) {
						sql.append(", ");
					}
				}
				return sql.toString();
			}
		}

	}

	/**
	 * Workaround for Hibernate has a bug https://hibernate.onjira.com/browse/HHH-6915
	 * 
	 */
	private static class ILikeExpression extends LikeExpression {

		private static final long serialVersionUID = 200053570285372152L;

		public ILikeExpression(String propertyName, String value) {
			super(propertyName, value, null, true);
		}

	}

	/**
	 * Needed for cglib usage ONLY
	 */
	HibernateRepository() {
		this(null, null);// allows us to use final with fields despite CGLib usage
	}
	
	static Partition pageToPartition(Page page){
		return new Partition((page.getNumber()-1)*page.getSize(), page.getSize());
	}

}
